pub const DATETIME_AS_NANOS: &str = "https://github.com/DataTreehouse/chrontext#DateTimeAsNanos";
pub const NANOS_AS_DATETIME: &str = "https://github.com/DataTreehouse/chrontext#NanosAsDateTime";
pub const DATETIME_AS_SECONDS: &str =
    "https://github.com/DataTreehouse/chrontext#DateTimeAsSeconds";
pub const SECONDS_AS_DATETIME: &str =
    "https://github.com/DataTreehouse/chrontext#SecondsAsDateTime";

pub const FLOOR_DATETIME_TO_SECONDS_INTERVAL: &str =
    "https://github.com/DataTreehouse/chrontext#FloorDateTimeToSecondsInterval";
pub const MODULUS: &str = "https://github.com/DataTreehouse/chrontext#modulus";

pub const TIME_BUCKET: &str = "https://github.com/DataTreehouse/chrontext#time_bucket";
